/* -*- mode: c -*- */

/* CHARINT */

/*
 * default typemap.i dont allow pass integers 
 * to char type parameters. 
 * 
 * to use this workaround rename char to CHARINT 
 * and unsigned char to UCHARINT 
 * in function which need accept integers.
 * 
 * so, if original function definition look like
 * 
 *   void foo(char bar, unsigned char baz);
 * 
 * and it defined in .i file as
 * 
 *   void foo(CHARINT bar, UCHARINT baz);
 * 
 * then calls
 * 
 *   (foo 1, 2)        ; => correct
 *   (foo #\a, #\b)    ; => is also correct
 *   (foo 1, -1)       ; => not correct because second 
 *                          argument may not be negative number
 * 
 */

%inline %{
  typedef char CHARINT;
%}
%typemap(guile,in) CHARINT, const CHARINT
{
  if(SCM_INUMP($source))
  {
	$target = gh_scm2int($source);
  }
  else
  {
	$target = gh_scm2char($source);
  }
}
%typemap(guile,indoc) CHARINT, const CHARINT "($arg <integer>|<char>)"

%inline %{
  typedef unsigned char UCHARINT;
%}
%typemap(guile,in) UCHARINT, const UCHARINT
{
  if(SCM_INUMP($source))
  {
	$target = gh_scm2ulong($source);
  }
  else
  {
	$target = gh_scm2char($source);
  }
}
%typemap(guile,indoc) UCHARINT, const UCHARINT "($arg <integer>|<char>)"

/* OUTSET_BUF */

/*
 * set of typemaps for passing pointer to uniform vectors
 * and strings to C-side for modification by actual wrapped function.
 * 
 * so, if you want wrap function:
 * 
 *   void foo(char *str)
 *   {
 *     strcpy(str, "some text");
 *   };
 * 
 * declare it for swig using typemap SETBUF as:
 * 
 *   %name("foo!") extern void foo(char *SETBUF);
 * 
 * and then you can call it:
 * 
 *   (define str (make-string 128))
 *   (foo! str)
 *   str                    ; => "some text"
 *   (string-length str)    ; => 128
 * 
 * in previous example scheme variable str have length 128
 * but string "some text" have length 9. so:
 * 
 *   (string=? str "some text")   ; => #f
 * 
 * to prevent this behavior also presented special typemap 
 * SETBUF0 which change size of string on output to 
 * actual size of C string:
 * 
 *   %name("foo!") extern void foo(char *SETBUF0);
 * 
 *   (define str (make-string 128))
 *   (foo! str)
 *   str                          ; => "some text"
 *   (string-length str)          ; => 9
 *   (string=? str "some text")   ; => #t
 * 
 * SETBUF0 defined for char* and unsigned char* C-types.
 * SETBUF defined for char*, short*, long*, long long* 
 * and their unsigned equivalens. also macros 
 * SETBUF_MAP(C_NAME, SCM_NAME) provided for expandind
 * type definitions.
 * 
 * attention! C-side function know nothing about buffer boundaries.
 * you must support such info to prevent SIGSEGV-s.
 * 
 */

%typemap(guile,in) char *SETBUF, unsigned char *SETBUF
{
  if(SCM_IMP($source))
  {
	scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
  };
  switch(SCM_TYP7($source))
  {
   case scm_tc7_string:
	$target = SCM_STRING_CHARS($source);
	break;
   case scm_tc7_byvect:
	$target = SCM_UVECTOR_BASE($source);
	break;
   default:
	scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
	break;
  };
}
%typemap(guile,freearg) char *SETBUF, unsigned char *SETBUF "";

/* special variant for zero-terminated strings */
%typemap(guile,in) char *SETBUF0, unsigned char *SETBUF0
{
  if(SCM_NIMP($source) && SCM_STRINGP($source))
  {
	$target = SCM_STRING_CHARS($source);
  }
  else
  {
	scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
  };
}
%typemap(guile,argout) char *SETBUF0, unsigned char *SETBUF0
{
  size_t olen = SCM_STRING_LENGTH($source);
  size_t len = strlen($target);
  
  if(len >= olen)
  {
	scm_misc_error(FUNC_NAME, "Buffer out of bound", SCM_BOOL_F);
  }
  else
  {
	SCM_SET_STRING_CHARS($source, 
						 scm_must_realloc($target, olen, len + 1, FUNC_NAME));
	SCM_SET_STRING_LENGTH($source, len);
  };
}
%typemap(guile,freearg) char *SETBUF0, unsigned char *SETBUF0 "";

%define SETBUF_MAP(C_NAME, SCM_NAME)
  %typemap(guile,in) C_NAME *SETBUF
  {
	if(SCM_NIMP($source) && SCM_TYP7($source) == SCM_NAME)
	{
	  $target = SCM_UVECTOR_BASE($source);
	}
	else
	{
	  scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
	};
  }
  %typemap(guile,freearg) C_NAME *SETBUF "";
%enddef

SETBUF_MAP(short, scm_tc7_svect);
SETBUF_MAP(int, scm_tc7_ivect);
SETBUF_MAP(long, scm_tc7_ivect);
SETBUF_MAP(long long, scm_tc7_llvect);
SETBUF_MAP(unsigned short, scm_tc7_svect);
SETBUF_MAP(unsigned int, scm_tc7_uvect);
SETBUF_MAP(unsigned long, scm_tc7_uvect);
SETBUF_MAP(unsigned long long, scm_tc7_llvect);

/* SETBUFN & SETBUF_LEN */

/* 
 * set of typemaps for passing pointer to uniform vectors
 * and strings to C-side with buffer (string or vector) size info.
 * 
 * a lot of C functions accept pointer to buffer and it`s size.
 * if you want wrap function:
 * 
 *   void my_strncpy(char* dst, char *src, int len)
 *   {
 *     strncpy(dst, src, len);
 *   }
 * 
 * declare it for swig with SETBUFN and SETBUF_LEN typemaps:
 * 
 *   %name("my-strncpy!") 
 *     extern void my_strncpy(char *SETBUFN, char *src, 
 *                            int SETBUF_LEN);
 * 
 * then my-strncpy! can be called as:
 * 
 *   (define buf (make-uniform-array #\a 255))
 *   (my-strncpy! buf "hello")
 * 
 * and value of parameter len will be calculated as length of
 * uniform vector in _count_ of _units_.
 * 
 * Sometime is usefull to restrict length passed to function
 * by hand. For this reason SETBUF_OPTLEN typemap provided.
 * It`s act as SETBUF_LEN with one diffirence -- 
 * it is visible from scheme side:
 * 
 *   %name("my-strncpy!")
 *     extern void my_strncpy(char *SETBUFN, char *src,
 *                            int SETBUF_OPTLEN);
 * 
 * can be called as:
 * 
 *   (my-strncpy! buf "hello" 4)
 *   (display buf)                  ; => "hell"
 * 
 * In that case len parameter will be calculated as _minimum_
 * of length in _count_ of _units_ and supplied value.
 * 
 * Also, #f can be passed instread of numeric -- mean
 * pass to actual function length of buffer as SETBUF_LEN.
 * 
 * For \0 ended strings provided SETBUF0N 
 * (see note about SETBUF0 before).
 * 
 * For C-functions which accept two pairs of buffer and length
 * arguments -- typemaps SETBUFN_1 and SETBUF_[OPT]LEN_1 may
 * be used:
 * 
 *   void bar(int *buf1, int *buf2, int buf2_len, int buf1_len);
 * 
 * can be propogated to swig as:
 * 
 *   %name("bar!") 
 *     extern void bar(int *SETBUFN, int *SETBUFN_1,
 *                     int SETBUF_LEN, int SETBUF_OPTLEN_1);
 * 
 * SETBUF0N and SETBUF0N_1 defined for char* and 
 * unsigned char* C-types. SETBUFN and SETBUFN_1 
 * defined for char*, short*, long*, long long* 
 * and their unsigned equivalens. also macroses
 * SETBUFN_MAP(C_NAME, SCM_NAME) and 
 * SETBUFN_MAP_1(C_NAME, SCM_NAME) are provided 
 * for expandind type definitions.
 * 
 * Yet one attention: length of buffer passed to function
 * is in a _count_ of _elements_.
 * 
 */

%typemap(guile,ignore) int SETBUF_LEN, 
  unsigned int SETBUF_LEN,
  long SETBUF_LEN,
  unsigned long SETBUF_LEN,
  size_t SETBUF_LEN
{
  $target = 0;
  setbuf_len = &$target;
}

%typemap(guile,arginit) int SETBUF_OPTLEN, 
  unsigned int SETBUF_OPTLEN,
  long SETBUF_OPTLEN,
  unsigned long SETBUF_OPTLEN,
  size_t SETBUF_OPTLEN
{
  $target = 0;
  setbuf_len = &$target;
}
%typemap(guile,in) int SETBUF_OPTLEN, 
  unsigned int SETBUF_OPTLEN,
  long SETBUF_OPTLEN,
  unsigned long SETBUF_OPTLEN,
  size_t SETBUF_OPTLEN "";
%typemap(guile,check) int SETBUF_OPTLEN, 
  unsigned int SETBUF_OPTLEN,
  long SETBUF_OPTLEN,
  unsigned long SETBUF_OPTLEN,
  size_t SETBUF_OPTLEN
{
  /* FIXME: work only if *setbuf_len initialized earlier 
   * (by previous args)
   */
  if(SCM_NFALSEP($source) && gh_scm2ulong($source) <= *setbuf_len)
  {
	$target = gh_scm2ulong($source);
  }
}

%typemap(guile,in) char *SETBUFN(int *setbuf_len), 
                   unsigned char *SETBUFN(int *setbuf_len)
{
  if(SCM_IMP($source))
  {
	scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
  };
  switch(SCM_TYP7($source))
  {
   case scm_tc7_string:
	$target = SCM_STRING_CHARS($source);
	*setbuf_len = SCM_STRING_LENGTH($source);
	break;
   case scm_tc7_byvect:
	$target = SCM_UVECTOR_BASE($source);
	*setbuf_len = SCM_UVECTOR_LENGTH($source);
	break;
   default:
	scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
	break;
  };
}
%typemap(guile,freearg) char *SETBUFN, unsigned char *SETBUFN "";

%typemap(guile,in) char *SETBUF0N(int *setbuf_len), 
                   unsigned char *SETBUF0N(int *setbuf_len)
{
  if(SCM_NIMP($source) && SCM_STRINGP($source))
  {
	$target = SCM_STRING_CHARS($source);
	*setbuf_len = SCM_STRING_LENGTH($source);
  }
  else
  {
	scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
  };
}
%typemap(guile,argout) char *SETBUF0N, unsigned char *SETBUF0N
{
  size_t olen = SCM_STRING_LENGTH($source);
  size_t len = strlen($target);
  SCM_SET_STRING_CHARS($source, 
					   scm_must_realloc($target, olen, len + 1, FUNC_NAME));
  SCM_SET_STRING_LENGTH($source, len);
}
%typemap(guile,freearg) char *SETBUF0N, unsigned char *SETBUF0N "";

%define SETBUFN_MAP(C_NAME, SCM_NAME)
  %typemap(guile,in) C_NAME *SETBUFN(int *setbuf_len)
  {
	if(SCM_NIMP($source) && SCM_TYP7($source) == SCM_NAME)
	{
	  $target = SCM_UVECTOR_BASE($source);
	  *setbuf_len = SCM_UVECTOR_LENGTH($source);
	}
	else
	{
	  scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
	};
  }
  %typemap(guile,freearg) C_NAME *SETBUFN "";
%enddef

SETBUFN_MAP(short, scm_tc7_svect);
SETBUFN_MAP(int, scm_tc7_ivect);
SETBUFN_MAP(long, scm_tc7_ivect);
SETBUFN_MAP(long long, scm_tc7_llvect);
SETBUFN_MAP(unsigned short, scm_tc7_svect);
SETBUFN_MAP(unsigned int, scm_tc7_uvect);
SETBUFN_MAP(unsigned long, scm_tc7_uvect);
SETBUFN_MAP(unsigned long long, scm_tc7_llvect);

%typemap(guile,ignore) int SETBUF_LEN_1, 
  unsigned int SETBUF_LEN_1,
  long SETBUF_LEN_1,
  unsigned long SETBUF_LEN_1,
  size_t SETBUF_LEN_1
{
  $target = 0;
  setbuf_len_1 = &$target;
}
%typemap(guile,arginit) int SETBUF_OPTLEN_1, 
  unsigned int SETBUF_OPTLEN_1,
  long SETBUF_OPTLEN_1,
  unsigned long SETBUF_OPTLEN_1,
  size_t SETBUF_OPTLEN_1
{
  $target = 0;
  setbuf_len_1 = &$target;
}
%typemap(guile,in) int SETBUF_OPTLEN_1, 
  unsigned int SETBUF_OPTLEN_1,
  long SETBUF_OPTLEN_1,
  unsigned long SETBUF_OPTLEN_1,
  size_t SETBUF_OPTLEN_1 "";
%typemap(guile,check) int SETBUF_OPTLEN_1, 
  unsigned int SETBUF_OPTLEN_1,
  long SETBUF_OPTLEN_1,
  unsigned long SETBUF_OPTLEN_1,
  size_t SETBUF_OPTLEN_1
{
  /* FIXME: work only if *setbuf_len_1 initialized earlier 
   * (by previous args)
   */
  if(SCM_NFALSEP($source) && gh_scm2ulong($source) <= *setbuf_len_1)
  {
	$target = gh_scm2ulong($source);
  }
}
%typemap(guile,in) char *SETBUFN_1(int *setbuf_len_1), 
                   unsigned char *SETBUFN_1(int *setbuf_len_1)
{
  if(SCM_IMP($source))
  {
	scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
  };
  switch(SCM_TYP7($source))
  {
   case scm_tc7_string:
	$target = SCM_STRING_CHARS($source);
	*setbuf_len_1 = SCM_STRING_LENGTH($source);
	break;
   case scm_tc7_byvect:
	$target = SCM_UVECTOR_BASE($source);
	*setbuf_len_1 = SCM_UVECTOR_LENGTH($source);
	break;
   default:
	scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
	break;
  };
}
%typemap(guile,freearg) char *SETBUFN_1, 
                          unsigned char *SETBUFN_1 = char *SETBUFN;
%typemap(guile,in) char *SETBUF0N_1(int *setbuf_len_1), 
                   unsigned char *SETBUF0N_1(int *setbuf_len_1)
{
  if(SCM_NIMP($source) && SCM_STRINGP($source))
  {
	$target = SCM_STRING_CHARS($source);
	*setbuf_len_1 = SCM_STRING_LENGTH($source);
  }
  else
  {
	scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
  };
}
%typemap(guile,argout) char *SETBUF0N_1, 
                       unsigned char *SETBUF0N_1 = char *SETBUF0N;
%typemap(guile,freearg) char *SETBUF0N_1, 
                        unsigned char *SETBUF0N_1 = char *SETBUF0N;

%define SETBUFN_1_MAP(C_NAME, SCM_NAME)
  %typemap(guile,in) C_NAME *SETBUFN_1(int *setbuf_len_1)
  {
	if(SCM_NIMP($source) && SCM_TYP7($source) == SCM_NAME)
	{
	  $target = SCM_UVECTOR_BASE($source);
	  *setbuf_len_1 = SCM_UVECTOR_LENGTH($source);
	}
	else
	{
	  scm_wrong_type_arg(FUNC_NAME, $argnum, $source);
	};
  }
  %typemap(guile,freearg) C_NAME *SETBUFN_1 "";
%enddef

SETBUFN_1_MAP(short, scm_tc7_svect);
SETBUFN_1_MAP(int, scm_tc7_ivect);
SETBUFN_1_MAP(long, scm_tc7_ivect);
SETBUFN_1_MAP(long long, scm_tc7_llvect);
SETBUFN_1_MAP(unsigned short, scm_tc7_svect);
SETBUFN_1_MAP(unsigned int, scm_tc7_uvect);
SETBUFN_1_MAP(unsigned long, scm_tc7_uvect);
SETBUFN_1_MAP(unsigned long long, scm_tc7_llvect);
